# _*_ coding: utf-8 _*_
"""
Time:     2022-05-02 0:26
Author:   Haolin Yan(XiDian University)
File:     converter.py
"""
from .utils import *
from .parser import parse_yaml, parse_kargs
import logging
from tabulate import tabulate
import os
import pickle
import torch

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def convert2mnn(model, input_shape, path, verbose=False):
    """Convert pytorch model to mnn format."""
    model.eval().cuda()
    dummy_input = torch.randn(input_shape, device='cuda')
    input_names = ["input"]
    output_names = ["output"]
    assert isinstance(input_shape, list), "Invalid input_shape: {}".format(input_shape)
    name = os.path.join(path, replace(model.__repr__())) + shape2str(input_shape)
    onnx_name = name + ".onnx"
    mnn_name = name + ".mnn"
    try:
        torch.onnx.export(model, dummy_input, onnx_name, verbose=verbose, input_names=input_names,
                          output_names=output_names)
    except:
        logger.warning("%s maybe is a invalid ops!" % (onnx_name))
        return 0
    logger.info("Export %s to %s" % (name, onnx_name))
    cmd = "MNNConvert -f ONNX --modelFile {} --MNNModel {} --bizCode biz".format(onnx_name, mnn_name)
    os.system(cmd)
    if not os.path.isfile(mnn_name):
        raise ValueError("Fail to convert %s to %s" % (onnx_name, mnn_name))
    logger.info("Convert %s to %s" % (onnx_name, mnn_name))
    os.remove(onnx_name)
    return os.path.basename(mnn_name)


def generate_ops_list(path, fp):
    """
    Generate operators in mnn format according to the specified YAML file.
    """
    operator_list = parse_yaml(path)
    logger.info("Begin to convert %d ops!" % (len(operator_list)))
    if not os.path.isdir(fp):
        os.mkdir(fp)
        logger.info("Create %s" % fp)
    success_ops = []
    for i, ops in enumerate(operator_list):
        module, input_shape = ops.return_instance(), ops.input_shape
        mnn_name = convert2mnn(module, input_shape, fp, verbose=False)
        if mnn_name != 0:
            ops.record_mnn_fname(mnn_name)
            success_ops.append(ops)
    logger.info(
        "Conversion completed (%d successes, %d failures)" % (len(success_ops), len(operator_list) - len(success_ops)))
    meta_fpath = os.path.join(fp, "meta.pkl")
    with open(meta_fpath, "wb") as f:
        pickle.dump(success_ops, f)


def register(ops, fp, **kwargs):
    operator_list = parse_kargs(ops, kwargs)
    logger.info("Begin to convert %d ops!" % (len(operator_list)))
    if not os.path.isdir(fp):
        os.mkdir(fp)
        logger.info("Create %s" % fp)
    success_ops = []
    for i, ops in enumerate(operator_list):
        module, input_shape = ops.return_instance(), ops.input_shape
        mnn_name = convert2mnn(module, input_shape, fp, verbose=False)
        if mnn_name != 0:
            ops.record_mnn_fname(mnn_name)
            success_ops.append(ops)
    logger.info(
        "Conversion completed (%d successes, %d failures)" % (len(success_ops), len(operator_list) - len(success_ops)))
    meta_fpath = os.path.join(fp, "meta.pkl")
    if os.path.isfile(meta_fpath):
        with open(meta_fpath, "rb") as f:
            last_ops = pickle.load(f)
        success_ops += last_ops
    with open(meta_fpath, "wb") as f:
        pickle.dump(success_ops, f)


# Tools for exporting models in batches and testing them
def export_models(net, input_shape, fp, verbose=False):
    if not os.path.isdir(fp):
        os.mkdir(fp)
        logger.info("Create %s" % fp)
    torch_path = os.path.join(fp, net.__repr__()+'.pth')
    torch.save(net, torch_path)
    mnn_name = convert2mnn(net, input_shape, fp, verbose=verbose)
    kwargs = dict()
    kwargs["torch_name"] = net.__repr__()+'.pth'
    kwargs["input_shape"] = input_shape
    with open(os.path.join(fp, mnn_name[:-4] + "meta.json"), "w") as f:
        json.dump(kwargs, f)


def validation(fp, ops_path, verbose=False, save_path=None):
    from .parser import predict_latency
    headers = ["model", "latency_true(ms)", "latency_pred(ms)", "error(ms)"]
    data = []
    fnames = [x for x in os.listdir(fp) if x[-5:] == ".json"]
    for name in fnames:
        json_path = os.path.join(fp, name)
        with open(json_path, "r") as f:
            kargs = json.load(f)
            torch_name = kargs["torch_name"]
            latency_t = kargs["latency"]["Avg"]
            input_shape = kargs["input_shape"]
            net = torch.load(os.path.join(fp, torch_name))
            latency_p = predict_latency(net, ops_path, input_shape, verbose=verbose)
            if verbose:
                logger.info("{}: latency: {} pred_latency: {}".format(torch_name, latency_t, latency_p))
            data.append([torch_name, latency_t, latency_p, latency_p - latency_t])
    print(tabulate(data, headers))
    if save_path is not None:
        import pandas as pd
        import numpy as np
        data = np.array(data)
        pd.DataFrame({"model": data[:, 0], "latency_true(ms)": data[:, 1], "latency_pred(ms)": data[:, 2]}).to_csv(save_path)








# _*_ coding: utf-8 _*_
"""
Time:     2022-05-02 13:54
Author:   Haolin Yan(XiDian University)
File:     meter.py
"""
import json
import os
import re
import pickle
import logging


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def get_latency(path, times=30, verbose=True):
    """Test the latency of operator with the given path"""
    cmd = "MNNV2Basic.out %s %d 0 0 4 > %s" % (path, times, path + '.log')
    os.system(cmd)
    with open(path + '.log', 'r') as f:
        lines = f.readlines()
    if verbose:
        print(lines)
    logger.info("%s:::%s" % (path, lines[-1]))
    prefix = ["Avg", "Min", "Max"]
    result = dict((prefix[i], float(v)) for i, v in enumerate(re.findall(r"\d+\.?\d*", lines[-1])))
    result["repeat"] = times
    return result


def meter_ops(fp, times=30, verbose=False):
    """Test the latency of operators in the input folder, and write the result in meta_latency.pkl"""
    logger.info("Begin to measure!")
    meta_path = os.path.join(fp, "meta.pkl")
    with open(meta_path, "rb") as f:
        meta_list = pickle.load(f)
    for meta in meta_list:
        path = os.path.join(fp, meta.mnn_fname)
        result = get_latency(path, times=times, verbose=verbose)
        meta.record_mnn_performance(result)
    update_meta_path = os.path.join(fp, "meta_latency.pkl")
    with open(update_meta_path, "wb") as f:
        pickle.dump(meta_list, f)
    logger.info("Finish!")


def get_model_latency(model, input_shape, path=".", times=30):
    """
    Convert pytorch model to mnn format and test its latency
    """
    if len(model.__repr__()) > 50:
        logger.warn("please rewrite the model.__repr__()")
    from .converter import convert2mnn
    mnn_name = convert2mnn(model, input_shape, path, verbose=False)
    path = os.path.join(path, mnn_name)
    result = get_latency(path, times=times)
    os.remove(path)
    return result


def meter_models(fp, times=30, verbose=False):
    """Test the latency of model in the input folder, and write the result in *meta.json"""
    logger.info("Begin to measure!")
    fnames = [x for x in os.listdir(fp) if x[-3:] == "mnn"]
    for name in fnames:
        fpath = os.path.join(fp, name)
        result = get_latency(fpath, times=times, verbose=verbose)
        json_path = os.path.join(fp, name[:-4] + "meta.json")
        with open(json_path, "r") as f:
            kargs = json.load(f)
        kargs["latency"] = result
        with open(json_path, "w") as f:
            json.dump(kargs, f)
    logger.info("Finish!")


# _*_ coding: utf-8 _*_
"""
Time:     2022-05-02 13:20
Author:   Haolin Yan(XiDian University)
File:     parser.py
"""
from .utils import (import_module,
                    generate_param_list,
                    ops_meta,
                    parse_ops_info,
                    get_net_device)
import yaml
from tabulate import tabulate
import logging
import torch

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def predict_latency(module, ops_path, input_shape, verbose=False):
    """Predict the model's latency using the latency table"""
    headers = ["ops", "input_shape", "output_shape", "Avg latency(ms)"]
    ops_list = parse_ops_info(ops_path)
    result = []
    latency_all = [0]

    def get_input_shape(m, in_f, out_f):
        input_shape = list(in_f[0].shape)
        found = False
        for ops in ops_list:
            if isinstance(m, ops[0]) and input_shape == ops[1] and m.__repr__() == ops[-1]:
                result.append([ops[-1], ops[1], list(out_f.shape), ops[2]])
                latency_all[0] += ops[2]
                found = True
                break

        if not found:
            logger.warning("No matching operator is found, check whether the input format of the defined operator is "
                           "consistent with that in the operator description file")
            logger.warning("ops: %s \n ops_list:" % m.__repr__())
            for ops in ops_list:
                print(ops[-1])

    def search_ops(module, ops_list):
        for name, m in module.named_children():
            found = False
            for ops in ops_list:
                if isinstance(m, ops[0]):
                    if verbose:
                        print("{} belongs to {}".format(name, ops[0].__name__))
                    m.register_forward_hook(get_input_shape)
                    m.__name__ = name
                    found = True
                    break
            if not found:
                search_ops(m, ops_list)
        return module

    module = search_ops(module, ops_list)
    device = get_net_device(module)
    module.eval()
    x = torch.ones(input_shape).to(device)
    _ = module(x)
    result.append(("Sum", "-", "-", latency_all[0]))
    if verbose:
        print(tabulate(result, headers=headers))
    return latency_all[0]


def parse_kargs(module, param):
    ops_register = []
    if param.get("no_params", False):
        for in_shape in param["input_shape"]:
            ops_register.append(ops_meta(module, in_shape))
        return ops_register

    cfg_list = generate_param_list(param)
    for cfg in cfg_list:
        input_shape = cfg["input_shape"]
        del cfg["input_shape"]
        ops_register.append(ops_meta(module, input_shape, init_param=cfg))
    return ops_register


def parse_yaml(path):
    """
    Parse the YAML file and return all operators(a list of class)
    e.g.
        parse_yaml("docs/demo.yaml")
    """
    ops_register = []
    with open(path, encoding='utf-8') as f:
        data = yaml.load(f.read(), Loader=yaml.FullLoader)
    for k, v in data.items():
        if v == "None":
            continue
        # import package
        pkg = import_module(k)
        for ops, param in v.items():
            if param.get("no_params", False):
                for in_shape in param["input_shape"]:
                    ops_register.append(ops_meta(getattr(pkg, ops), in_shape))
                continue
            cfg_list = generate_param_list(param)
            for cfg in cfg_list:
                input_shape = cfg["input_shape"]
                del cfg["input_shape"]
                ops_register.append(ops_meta(getattr(pkg, ops), input_shape, init_param=cfg))
    return ops_register


def summary_model(model, input_size, ops_list, verbose=False):
    ops_input_shape = []
    headers = ["ops", "input_shape", "out_shape"]

    def get_input_shape(m, in_f, out_f):
        for ops in ops_list:
            if isinstance(m, ops):
                ops_input_shape.append([m.__repr__(), list(in_f[0].shape), list(out_f.shape)])
                break

    def search_ops(module, ops_list):
        for name, m in module.named_children():
            found = False
            for ops in ops_list:
                if isinstance(m, ops):
                    if verbose:
                        print("{} belongs to {}".format(name, ops.__name__))
                    m.register_forward_hook(get_input_shape)
                    m.__name__ = name
                    found = True
                    break
            if not found:
                search_ops(m, ops_list)
        return module

    module = search_ops(model, ops_list)
    device = get_net_device(module)
    module.eval()
    x = torch.ones(input_size).to(device)
    _ = module(x)
    print(tabulate(ops_input_shape, headers=headers))



# _*_ coding: utf-8 _*_
"""
Time:     2022-05-01 17:23
Author:   Haolin Yan(XiDian University)
File:     utils.py
"""
import importlib
import copy
import pickle
import os
import json
from .meter import get_latency


class ops_meta:
    def __init__(self, classesname, input_shape, init_param=None):
        self.classname = classesname
        self.input_shape = input_shape
        self.init_param = init_param
        self.mnn_fname = None
        self.avg = 0
        self.max = 0
        self.min = 0
        self.repr = ""

    def __repr__(self):
        return self.repr

    def return_instance(self):
        if self.init_param:
            model = self.classname(**self.init_param)
        else:
            model = self.classname()
        self.repr = model.__repr__()
        return model

    def record_mnn_fname(self, fname):
        self.mnn_fname = fname

    def record_mnn_performance(self, result):
        self.avg = result["Avg"]
        self.min = result["Min"]
        self.max = result["Max"]


def import_module(pkg):
    return importlib.import_module(pkg)


def shape2str(shape):
    s = ""
    for i in shape:
        s += str(i) + '-'
    return s[:-1]


def generate_param_list(kwargs):
    cfg_list = []
    index = 0
    key_list = list(kwargs.keys())

    def generate_param(param, index, **kwargs):
        # print(param, index)
        key = key_list[index]
        if key == "input_shape":
            cfg_list.append(param)
            return 0

        table = kwargs[key]
        for j, ele in enumerate(table):
            param[key] = ele
            param_ = copy.deepcopy(param)
            if index == 0:
                param_["input_shape"] = kwargs["input_shape"][j]
            generate_param(param_, index + 1, **kwargs)

    param = dict((n, None) for n in kwargs.keys())
    generate_param(param, index, **kwargs)
    return cfg_list


def remove_(str):
    last = ""
    new_str = ""
    for s in str:
        if s == last:
            continue
        new_str += s
        last = s
    return new_str


def replace(str):
    str = str.replace("(", "-").replace(")", "-").replace(",", "-").replace(" ", "")
    return remove_(str)


def parse_ops_info(path):
    ops_info_list = []
    with open(path, "rb") as f:
        ops_list = pickle.load(f)
        for ops in ops_list:
            ops_info_list.append((ops.classname, ops.input_shape, ops.avg, ops.min, ops.max, ops.repr))
    return ops_info_list


def get_net_device(model):
    return next(model.parameters()).device


def export_models(model, input_shape, path, **kwargs):
    from .converter import convert2mnn
    net = model(**kwargs)
    convert2mnn(net, input_shape, path)
    kwargs["fname"] = path
    with open(os.path.join(path, "meta.json"), "w") as f:
        json.dump(kwargs, f)











# _*_ coding: utf-8 _*_
"""
Time:     2022-05-04 1:10
Author:   Haolin Yan(XiDian University)
File:     setup.py.py
"""
from setuptools import setup, find_packages

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='mnn-meter',
    include_package_data=True,
    version="2.0.0",
    author="Haolin Yan",
    author_email='haolinyan_xdu@163.com',
    description="Tools for quickly building operator latency tables and for accurately predicting model latency (based on Pytorch and MNN)",
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'Programming Language :: Python :: 3',
        # 省略一下
    ],
    packages=find_packages(),
    python_requires='>=3.5',
    install_requires=[
        "tabulate",
        "pyyaml",
        "pandas",
        "numpy"
    ],
    project_urls={
        'Source': 'https://github.com/makerlin1/MMT',
    }
)
